import numpy as np
from collections import namedtuple

def get_top_k_indexes(x: np.ndarray, k: int) -> np.ndarray:
    topk = np.argpartition(x, -k)[-k:]
    return topk


TopK = namedtuple("TopK", ["values", "index"])

def get_top_k(x: np.ndarray, k:int) -> TopK:
    idxs = get_top_k_indexes(x, k)
    topk = x[idxs]
    sorted_topk_idxs = np.argsort(-topk)
    return TopK(topk[sorted_topk_idxs], idxs[sorted_topk_idxs])


import numpy as np
from forestknn.utils import *

def matching_leaf_indexes(x, leaf_idx):
    return (x == leaf_idx).astype(int)

def get_matching_leaf_indexes(xs, leaf_idxs):
    matches = list(map(lambda x: matching_leaf_indexes(*x), zip(xs.T, leaf_idxs.T)))
    return np.vstack(matches)

def calculate_neighbor_weights(match_idxs):
    return match_idxs.sum(axis=0)/match_idxs.sum()

def neighbor_weights(estimator, neighborhood, query):
    neighborhood_leaf_idxs = estimator.apply(neighborhood)
    query_leaf_idx = estimator.apply(query)
    matches = get_matching_leaf_indexes(neighborhood_leaf_idxs, query_leaf_idx)
    wts = calculate_neighbor_weights(matches)
    return wts

def adaptive_k_nearest_neighbors(estimator, neighborhood, query, k = 10):
    neighbor_wts = neighbor_weights(estimator, neighborhood, query)
    top_k = get_top_k(neighbor_wts, k)
    knns = neighborhood.iloc[top_k.index].assign(wts = lambda x: top_k.values)
    return knns